from pyuarm import *
from list_uarms import uarm_ports
import copy

INIT_POS_L = 139
INIT_POS_R = 26
SAMPLING_DEADZONE = 2


class Calibration(object):
    manual_calibration_trigger = True
    linear_calibration_start_flag = False
    stretch_calibration_flag = False

    linear_offset_template = {"INTERCEPT": 0.00, "SLOPE": 0.00}
    linear_offset = [linear_offset_template, linear_offset_template, linear_offset_template, linear_offset_template]
    linear_offset_servo_flag = [False, False, False, False]
    manual_offset = [0.00, 0.00, 0.00]
    stretch_offset_template = {"LEFT": 0.00, "RIGHT": 0.00}
    temp_manual_offset_arr = [0.00, 0.00, 0.00]
    manual_offset_correct_flag = [False, False, False]

    def __init__(self, uarm=None, log_function=None):
        if uarm is not None:
            self.uarm = uarm
        elif len(uarm_ports()) > 0:
            self.uarm = get_uarm()
        else:
            raise NoUArmPortException('No available uArm Founds')

        if log_function is not None:
            self.log_function = log_function
        else:
            self.log_function = self.default_log
        self.servo_calibrate_timeout = 300
        self.is_all_calibrated = False
        self.is_linear_calibrated = False
        self.is_manual_calibrated = False
        self.is_stretch_calibrated = False

    @staticmethod
    def default_log(msg):
        print msg

    def uf_print(self, msg):
        self.log_function(msg)

    def calibrate_all(self, linear_callback=None, manual_callback=None, stretch_callback=None):
        self.uf_print("0. Clearing Completed Flag in EEPROM.")
        self.write_completed_flag(CALIBRATION_FLAG, False)

        if linear_callback is None:  # If not provide callback function use default
            self.linear_calibration_section(None)
        else:
            linear_callback()

        if self.read_completed_flag(CALIBRATION_LINEAR_FLAG):
            if manual_callback is None:  # If not provide callback function use default
                self.manual_calibration_section(None)
            else:
                manual_callback()
            time.sleep(0.5)

        if self.read_completed_flag(CALIBRATION_SERVO_FLAG):
            if stretch_callback is None:  # If not provide callback function use default
                self.stretch_calibration_section(None)
            else:
                stretch_callback()
            time.sleep(0.5)

        if self.read_completed_flag(CALIBRATION_STRETCH_FLAG):
            self.write_completed_flag(CALIBRATION_FLAG, True)
            self.uf_print("Calibration DONE!!")

    def linear_calibration_section(self, callback=None):
        self.linear_calibration_start_flag = True
        self.uf_print("1.0. Clearing Linear Completed Flag in EEPROM.")
        self.write_completed_flag(CALIBRATION_LINEAR_FLAG, False)
        self.uf_print("1. Start Calibrate Linear Offset")
        for i in range(4):
            self.uf_print("    1.1. {0} Linear Offset - Servo {1}".format(str(i), str(i)))
            temp_linear_offset = self.calibrate_linear_servo_offset(i)
            linear_is_correct = False
            if check_linear_is_correct(temp_linear_offset):
                self.linear_offset[i] = temp_linear_offset
                linear_is_correct = True
            else:
                linear_is_correct = False
            if callback is not None:
                callback(i, linear_is_correct, temp_linear_offset)

        self.save_linear_offset()
        time.sleep(1)
        if self.read_linear_offset() == self.linear_offset:
            self.uf_print("    1.3 Mark Completed Flag in EEPROM")
            self.write_completed_flag(CALIBRATION_LINEAR_FLAG, True)
            self.linear_calibration_start_flag = False
        else:
            self.uf_print("Error - 1. Linear Offset not equal to EEPROM, Please retry.")
            self.uf_print("Error - 1. linear_offset(): {0}".format(self.read_linear_offset()))
            self.linear_calibration_start_flag = False
        self.uarm.detach_all_servos()

    def calibrate_linear_servo_offset(self, number):
        global analog_read_pin, servo_analog_read
        moveTimes = 5
        ServoRangeIni = 20
        ServoRangeFin = 100
        angle_step = 0

        angles = []
        analogs = []
        self.ab_values = []
        while angle_step < ((ServoRangeFin - ServoRangeIni) / moveTimes + 1):
            angle = angle_step * moveTimes + ServoRangeIni
            if number == SERVO_BOTTOM:
                analog_read_pin = SERVO_BOTTOM_ANALOG_PIN
                self.uarm.write_servo_angle(SERVO_BOTTOM, angle, 0)
                self.uarm.write_left_right_servo_angle(60, 30, 0)

            if number == SERVO_LEFT:
                analog_read_pin = SERVO_LEFT_ANALOG_PIN
                self.uarm.write_servo_angle(SERVO_BOTTOM, 90, 0)
                self.uarm.write_left_right_servo_angle(angle, 30, 0)

            if number == SERVO_RIGHT:
                analog_read_pin = SERVO_RIGHT_ANALOG_PIN
                self.uarm.write_servo_angle(SERVO_BOTTOM, 90, 0)
                self.uarm.write_left_right_servo_angle(30, angle, 0)

            if number == SERVO_HAND:
                analog_read_pin = SERVO_HAND_ANALOG_PIN
                self.uarm.write_servo_angle(SERVO_BOTTOM, 90, 0)
                self.uarm.write_left_right_servo_angle(30, 60, 0)
                self.uarm.write_servo_angle(SERVO_HAND, angle, 0)

            if angle_step == 0:
                time.sleep(0.1)
            else:
                time.sleep(0.1)

            for i in range(2):
                servo_analog_read = self.uarm.read_analog(analog_read_pin)
                time.sleep(0.1)

            angles.append(angle)
            analogs.append(servo_analog_read)
            angle_step += 1
            time.sleep(0.1)

        new_ab = basic_linear_regression(analogs, angles)
        linear_offset_template = copy.deepcopy(self.linear_offset_template)
        linear_offset_template['SLOPE'] = round(new_ab[0], 2)
        linear_offset_template['INTERCEPT'] = round(new_ab[1], 2)
        return linear_offset_template

    def manual_calibration_section(self, callback=None):
        self.uf_print("2.0. Clearing Servo Completed Flag in EEPROM.")
        self.write_completed_flag(CALIBRATION_SERVO_FLAG, False)
        self.uf_print("2. Start Calibrate Servo Offset")
        self.manual_operation_trigger = True
        self.uarm.write_servo_angle(SERVO_BOTTOM, 45, 0)
        time.sleep(1)
        self.uarm.write_left_right_servo_angle(130, 20, 0)
        time.sleep(1)
        self.uarm.detach_all_servos()
        time.sleep(0.3)

        time_counts = 0

        servo_1_offset = 0
        servo_2_offset = 0
        servo_3_offset = 0
        self.uf_print("Please move uArm in right position")
        self.uarm.alarm(3, 100, 100)
        while self.manual_operation_trigger:

            servo_1_offset = self.uarm.read_servo_angle(SERVO_BOTTOM, 0) - 45
            servo_2_offset = self.uarm.read_servo_angle(SERVO_LEFT, 0) - 130
            servo_3_offset = self.uarm.read_servo_angle(SERVO_RIGHT, 0) - 20

            if abs(servo_1_offset) < 5.5:
                self.manual_offset_correct_flag[0] = True
            else:
                self.manual_offset_correct_flag[0] = False
                # self.uf_print("Please try move the Servo 1")
            if abs(servo_2_offset) < 5.5:
                self.manual_offset_correct_flag[1] = True
            else:
                self.manual_offset_correct_flag[1] = False
                # self.uf_print("Please try move the Servo 2")
            if abs(servo_3_offset) < 5.5:
                self.manual_offset_correct_flag[2] = True
            else:
                self.manual_offset_correct_flag[2] = False
                # self.uf_print("Please try move the Servo 3")

            if time_counts > self.servo_calibrate_timeout:
                self.manual_operation_trigger = False
            time_counts += 1
            self.temp_manual_offset_arr[0] = servo_1_offset
            self.temp_manual_offset_arr[1] = servo_2_offset
            self.temp_manual_offset_arr[2] = servo_3_offset
            if self.manual_offset_correct_flag[0] & self.manual_offset_correct_flag[1] & \
                    self.manual_offset_correct_flag[2]:
                self.uf_print(str(self.servo_calibrate_timeout - time_counts) + ", Please Confirm the positions")
            if callback is not None:
                callback(self.temp_manual_offset_arr, self.manual_offset_correct_flag)
            else:
                confirm = raw_input(
                    "servo offset, bottom: {0}, left: {1}, right: {2},\nConfirm Please Press Y: ".format(servo_1_offset,servo_2_offset,servo_3_offset))
                if confirm == "Y" or confirm == "y":
                    self.manual_operation_trigger = False
            time.sleep(0.1)

        self.temp_manual_offset_arr[0] = round(servo_1_offset, 2)
        self.temp_manual_offset_arr[1] = round(servo_2_offset, 2)
        self.temp_manual_offset_arr[2] = round(servo_3_offset, 2)
        self.save_manual_offset()
        if self.read_manual_offset() == self.temp_manual_offset_arr:
            self.uf_print("    2.3 Mark Completed Flag in EEPROM")
            self.write_completed_flag(CALIBRATION_SERVO_FLAG, True)
        else:
            self.write_completed_flag(CALIBRATION_SERVO_FLAG, False)
            self.uf_print("Error - 2, Manual Calibration Servo offset not equal to EEPROM")
            self.uf_print("Error - 2, manual Servo Offset: {0}".format(self.temp_manual_offset_arr))
            self.uf_print("Error - 2, read_manual_offset: {0}".format(self.read_manual_offset()))
        self.uarm.detach_all_servos()

    def stretch_calibration_section(self, callback=None):
        self.stretch_calibration_flag = True
        self.uf_print("3.0. Clearing Stretch Completed Flag in EEPROM.")
        self.write_completed_flag(CALIBRATION_STRETCH_FLAG, False)
        self.uf_print("3. Start Calibrate Stretch Offset")

        self.uf_print("3.0 Moving uArm to Correct Place")
        self.uarm.write_servo_angle(SERVO_BOTTOM, 45, 0)
        time.sleep(1)
        self.uarm.write_left_right_servo_angle(130, 20, 0)
        time.sleep(1)

        initPosL = INIT_POS_L - 12
        initPosR = INIT_POS_R - 12
        minAngle_L = self.uarm.read_analog(SERVO_LEFT_ANALOG_PIN) - 16;
        minAngle_R = self.uarm.read_analog(SERVO_RIGHT_ANALOG_PIN) - 16;
        print ('minAngle_L: {0}'.format(minAngle_L))
        print ('minAngle_R: {0}'.format(minAngle_R))

        self.uarm.write_servo_angle(SERVO_LEFT, initPosL, 0)
        self.uarm.write_servo_angle(SERVO_RIGHT, initPosR, 0)
        time.sleep(1)
        while self.uarm.read_analog(SERVO_RIGHT_ANALOG_PIN) < (minAngle_R - SAMPLING_DEADZONE) \
                and self.stretch_calibration_flag:
            initPosR += 1
            self.uarm.write_servo_angle(SERVO_RIGHT, initPosR, 0)
            print ('initPosR: {0}'.format(initPosR))
            time.sleep(0.05)

        while self.uarm.read_analog(SERVO_RIGHT_ANALOG_PIN) > (minAngle_R + SAMPLING_DEADZONE) \
                and self.stretch_calibration_flag:
            initPosR -= 1
            self.uarm.write_servo_angle(SERVO_RIGHT, initPosR, 0)
            print ('initPosR: {0}'.format(initPosR))
            time.sleep(0.05)

        while self.uarm.read_analog(SERVO_LEFT_ANALOG_PIN) < (minAngle_L - SAMPLING_DEADZONE) \
                and self.stretch_calibration_flag:
            initPosL += 1
            self.uarm.write_servo_angle(SERVO_LEFT, initPosL, 0)
            print ('initPosL: {0}'.format(initPosL))
            time.sleep(0.05)

        while self.uarm.read_analog(SERVO_LEFT_ANALOG_PIN) > (minAngle_L + SAMPLING_DEADZONE) \
                and self.stretch_calibration_flag:
            initPosL -= 1
            self.uarm.write_servo_angle(SERVO_LEFT, initPosL, 0)
            print ('initPosL: {0}'.format(initPosL))
            time.sleep(0.05)

        offsetL = initPosL - INIT_POS_L + 3
        offsetR = initPosR - INIT_POS_R + 3
        offset_correct_flag = [False, False]
        if abs(offsetL) < 20:
            offset_correct_flag[0] = True
        if abs(offsetR) < 20:
            offset_correct_flag[1] = True

        stretch_offset = copy.deepcopy(self.stretch_offset_template)
        stretch_offset['LEFT'] = offsetL
        stretch_offset['RIGHT'] = offsetR
        if callback is not None:
            callback(stretch_offset, offset_correct_flag)

        self.uf_print("    3.1 Saving Stretch Offset into EEPROM")
        self.uarm.write_eeprom(EEPROM_DATA_TYPE_FLOAT, OFFSET_STRETCH_START_ADDRESS, offsetL)
        self.uarm.write_eeprom(EEPROM_DATA_TYPE_FLOAT, OFFSET_STRETCH_START_ADDRESS + 4, offsetR)
        self.uf_print("    3.2 Mark Completed Flag in EEPROM")
        self.write_completed_flag(CALIBRATION_STRETCH_FLAG, True)
        print offsetL, offsetR
        self.uarm.detach_all_servos()
        self.stretch_calibration_flag = False

    def save_linear_offset(self):
        self.uf_print("    1.2 Saving Servo Offset into EEPROM")
        intercept_address = LINEAR_INTERCEPT_START_ADDRESS
        slope_address = LINEAR_SLOPE_START_ADDRESS
        for i in range(4):
            print ("Intercept Address:{0}, Offset Value:{1}.".format(intercept_address, self.linear_offset[i]['INTERCEPT']))
            print ("Slope Address:{0}, Offset Value:{1}.".format(slope_address, self.linear_offset[i]['SLOPE']))
            self.uarm.write_eeprom(EEPROM_DATA_TYPE_FLOAT, intercept_address, self.linear_offset[i]['INTERCEPT'])
            time.sleep(0.5)
            self.uarm.write_eeprom(EEPROM_DATA_TYPE_FLOAT, slope_address, self.linear_offset[i]['SLOPE'])
            time.sleep(0.5)
            intercept_address += 4
            slope_address += 4

    def save_manual_offset(self):
        self.uf_print("    2.1 Saving Servo Offset into EEPROM")
        address = OFFSET_START_ADDRESS
        for i in self.temp_manual_offset_arr:
            self.uarm.write_eeprom(EEPROM_DATA_TYPE_FLOAT, address, i)
            address += 4

    def read_manual_offset(self):
        address = OFFSET_START_ADDRESS
        read_manual_offset = []
        for i in range(3):
            read_manual_offset.append(round(self.uarm.read_eeprom(EEPROM_DATA_TYPE_FLOAT, address), 2))
            address += 4
        return read_manual_offset

    def read_linear_offset(self):
        intercept_address = LINEAR_INTERCEPT_START_ADDRESS
        slope_address = LINEAR_SLOPE_START_ADDRESS
        linear_offset_data = []
        for i in range(4):
            linear_offset_template = copy.deepcopy(self.linear_offset_template)
            linear_offset_template['INTERCEPT'] = round(
                self.uarm.read_eeprom(EEPROM_DATA_TYPE_FLOAT, intercept_address), 2)
            linear_offset_template['SLOPE'] = round(self.uarm.read_eeprom(EEPROM_DATA_TYPE_FLOAT, slope_address), 2)
            linear_offset_data.append(linear_offset_template)
            intercept_address += 4
            slope_address += 4
        return linear_offset_data

    def read_stretch_offset(self):
        address = OFFSET_STRETCH_START_ADDRESS
        left_offset = round(self.uarm.read_eeprom(EEPROM_DATA_TYPE_FLOAT, address), 2)
        right_offset = round(self.uarm.read_eeprom(EEPROM_DATA_TYPE_FLOAT, address + 4), 2)
        stretch_offset = copy.deepcopy(self.stretch_offset_template)
        stretch_offset['LEFT'] = left_offset
        stretch_offset['RIGHT'] = right_offset
        return stretch_offset

    def init_calibration_completed_flag(self):
        self.is_all_calibrated = self.read_completed_flag(CALIBRATION_FLAG)
        self.is_linear_calibrated = self.read_completed_flag(CALIBRATION_LINEAR_FLAG)
        self.is_manual_calibrated = self.read_completed_flag(CALIBRATION_SERVO_FLAG)
        self.is_stretch_calibrated = self.read_completed_flag(CALIBRATION_STRETCH_FLAG)

    def write_completed_flag(self, flag_type, flag):
        if flag_type == CALIBRATION_FLAG:
            self.uarm.write_eeprom(EEPROM_DATA_TYPE_BYTE, CALIBRATION_FLAG, CONFIRM_FLAG if flag else 0)
        elif flag_type == CALIBRATION_LINEAR_FLAG:
            self.uarm.write_eeprom(EEPROM_DATA_TYPE_BYTE, CALIBRATION_LINEAR_FLAG, CONFIRM_FLAG if flag else 0)
        elif flag_type == CALIBRATION_SERVO_FLAG:
            self.uarm.write_eeprom(EEPROM_DATA_TYPE_BYTE, CALIBRATION_SERVO_FLAG, CONFIRM_FLAG if flag else 0)
        elif flag_type == CALIBRATION_STRETCH_FLAG:
            self.uarm.write_eeprom(EEPROM_DATA_TYPE_BYTE, CALIBRATION_STRETCH_FLAG, CONFIRM_FLAG if flag else 0)

    def read_completed_flag(self, flag_type):
        if flag_type == CALIBRATION_FLAG:
            if self.uarm.read_eeprom(EEPROM_DATA_TYPE_BYTE, CALIBRATION_FLAG) == CONFIRM_FLAG:
                return True
            else:
                return False
        elif flag_type == CALIBRATION_LINEAR_FLAG:
            if self.uarm.read_eeprom(EEPROM_DATA_TYPE_BYTE, CALIBRATION_LINEAR_FLAG) == CONFIRM_FLAG:
                return True
            else:
                return False
        elif flag_type == CALIBRATION_SERVO_FLAG:
            if self.uarm.read_eeprom(EEPROM_DATA_TYPE_BYTE, CALIBRATION_SERVO_FLAG) == CONFIRM_FLAG:
                return True
            else:
                return False
        elif flag_type == CALIBRATION_STRETCH_FLAG:
            if self.uarm.read_eeprom(EEPROM_DATA_TYPE_BYTE, CALIBRATION_STRETCH_FLAG) == CONFIRM_FLAG:
                return True
            else:
                return False


def basic_linear_regression(x, y):
    def float_data(a):
        for i in range(len(a)):
            a[i] = float(a[i])
        return a

    x = float_data(x)
    y = float_data(y)

    # Basic computations to save a little time.
    length = len(x)
    sum_x = sum(x)
    sum_y = sum(y)

    sum_x_squared = sum(map(lambda a: a * a, x))
    sum_of_products = sum([x[i] * y[i] for i in range(length)])

    # Magic formulae!
    a = (sum_of_products - (sum_x * sum_y) / length) / (sum_x_squared - ((sum_x ** 2) / length))
    b = (sum_y - a * sum_x) / length
    return a, b


def check_linear_is_correct(linear_offset):
    if linear_offset['SLOPE'] > 0.1 and linear_offset['INTERCEPT'] < 1:
        return True


def main():
    """
    uArm Calibration Library
        1. Linear Calibration section
          1.1 Write 20 times Angle to each Servo, in the meanwhile read the 20 times Servo Analog
          1.2 Using basic_linear_regression(analogs, angles) to generate the intercept & slope
              ``(In arduino library, angle = analog * INTERCEPT + SLOPE)``
          1.3 Save each servo linear offset in EEPROM (LINEAR_INTERCEPT_START_ADDRESS, LINEAR_SLOPE_START_ADDRESS)
          1.4 Mark Linear Calibration in EEPROM (CALIBRATION_LINEAR_FLAG)
         2. Manual Calibration section
          2.1 Set uArm at standard Angles. Left Angle - 130, Right Angle 30, Bottom Angle, 45.
          2.2 Put uArm into the calibration position.
          2.3 Confirm the Position.
         3. Stretch Calibration section
    """
    calibration = Calibration()
    calibration.calibrate_all()


if __name__ == '__main__':
    main()



from tools.list_uarms import uarm_ports
import pyuarm

# Firmata

REPORT_FIRMATA_VERSION = 0x79

# parameters
SERVO_BOTTOM = 0
SERVO_LEFT = 1
SERVO_RIGHT = 2
SERVO_HAND = 3

SERVO_BOTTOM_ANALOG_PIN = 2
SERVO_LEFT_ANALOG_PIN = 0
SERVO_RIGHT_ANALOG_PIN = 1
SERVO_HAND_ANALOG_PIN = 3

# communication protocol
START_SYSEX = 0xF0
UARM_CODE = 0xAA
END_SYSEX = 0xF7

READ_ANGLE = 0X10
WRITE_ANGLE = 0X11
READ_COORDS = 0X12
WRITE_COORDS = 0X13
READ_DIGITAL = 0X14
WRITE_DIGITAL = 0X15
READ_ANALOG = 0X16
WRITE_ANALOG = 0X17
READ_EEPROM = 0X1A
WRITE_EEPROM = 0X1B
DETACH_SERVO = 0X1C
PUMP_STATUS = 0X1D
WRITE_STRETCH = 0X1E
WRITE_LEFT_RIGHT_ANGLE = 0X1F
GRIPPER_STATUS = 0X20
READ_SERIAL_NUMBER = 0x21
WRITE_SERIAL_NUMBER = 0x22
REPORT_FIRMWARE_VERSION = 0x23
BUZZER_ALERT = 0x24

CONFIRM_FLAG = 0x80
CALIBRATION_FLAG = 10
CALIBRATION_LINEAR_FLAG = 11
CALIBRATION_SERVO_FLAG = 12
CALIBRATION_STRETCH_FLAG = 13

LINEAR_INTERCEPT_START_ADDRESS = 70
LINEAR_SLOPE_START_ADDRESS = 50
OFFSET_START_ADDRESS = 30
OFFSET_STRETCH_START_ADDRESS = 20

SERIAL_NUMBER_ADDRESS = 100

EEPROM_DATA_TYPE_BYTE = 1
EEPROM_DATA_TYPE_INTEGER = 2
EEPROM_DATA_TYPE_FLOAT = 4

BUTTON_D7 = 7
BUTTON_D4 = 4
BUTTON_D2 = 2
PUMP_PIN = 6
VALVE_PIN = 5

PULL_UP = 1
INPUT = 0

HIGH = 1
LOW  = 0

ABSOLUTE = 0
RELATIVE = 1
PATH_LINEAR = 0   # path based on linear interpolation
PATH_ANGLES = 1   # path based on interpolation of servo angles

# interpolation types
INTERP_EASE_INOUT_CUBIC = 0  # original cubic ease in/out
INTERP_LINEAR           = 1
INTERP_EASE_INOUT       = 2  # quadratic easing methods
INTERP_EASE_IN          = 3
INTERP_EASE_OUT         = 4

def get_uarm():
    """
    ===============================
    Get First uArm Port instance
    ===============================
    It will return the first uArm Port detected by **pyuarm.tools.list_uarms**,
    If no available uArm ports, will print *There is no uArm port available*

    .. raw:python
    >>> import pyuarm
    >>> uarm = pyuarm.get_uarm()
    There is no uArm port available


    :returns: uArm() Instance

    """
    ports = uarm_ports()
    if len(ports) > 0:
        return pyuarm.uArm(port=ports[0])
    else:
        print("There is no uArm port available")


def getOne7BitBytesFloatArray(val):
    """
    :param val:

    """
    return bytearray([val])


def getTwo7BitBytesIntegerArray(val):
    """

    :param val:

    """
    int_val = int(val)
    return bytearray([abs(int_val) % 128, abs(int_val) >> 7])


def getThree7BitBytesFloatArray(val):
    """

    :param val:

    """
    int_val = int(val)
    decimal_val = int(round((val - int_val) * 100))
    return bytearray([abs(int_val) % 128, abs(int_val) >> 7, abs(decimal_val)])


def getThree7BitBytesIntegerArray(val):
    """

    :param val:

    """
    return bytearray([0 if val > 0 else 1, abs(val) % 128, abs(val) >> 7])


def getFour7BitBytesFloatArray(val):
    """

    :param val: Float Type
    :return: 7BitBytes Array, length = 4
    """
    int_val = int(val)
    decimal_val = int(round((val - int_val) * 100))
    return bytearray([0 if val > 0 else 1, abs(int_val) % 128, abs(int_val) >> 7, abs(decimal_val)])


def getFloatFromFour7BitBytes(array):
    """

    :param array: 4 bytes array
    :return: Float Type
    """
    negative = -1 if ord(array[0]) == 1 else 1
    val = negative * (ord(array[1]) + ord(array[2]) * 128 + ord(array[3]) / 100.00)
    return val

class NoUArmPortException(IOError):
    """
    .. _NoUArmPortException:
    """
    def __init__(self, error):
        self.error = error

    def __str__(self):
        return repr(self.error)


class UnkwonFirmwareException(Exception):
    """
    .. _UnkwonFirmwareException:
    """
    def __init__(self, error):
        self.error = error

    def __str__(self):
        return repr(self.error)



#!/usr/bin/env python
#
# This is a module that help user to manage the uArm firmware.
# Please use -h to list all parameters of this script

# This file is part of pyuarm. https://github.com/uArm-Developer/pyuarm
# (C) 2016 UFACTORY <developer@ufactory.cc>

import pyuarm
from pyuarm.tools.list_uarms import uarm_ports
import pycurl, certifi
import json
from io import BytesIO
from progressbar import ProgressBar, Percentage, FileTransferSpeed, Bar, ETA
import requests
import os, sys, platform, subprocess
from itertools import izip

from distutils.version import LooseVersion
import argparse

github_release_url = "https://api.github.com/repos/uArm-Developer/UArmForArduino/releases/latest"


class FirmwareHelper():
    firmware_defaul_filename = 'firmware.hex'
    application_path = ""
    firmware_path = ""
    web_firmware_version = ""
    uarm_firmware_version = ""
    uarm_port = ""

    def __init__(self):
        self.web_firmware_version = "0.0.0"
        if getattr(sys, 'frozen', False):
            self.application_path = os.path.dirname(sys.executable)
        elif __file__:
            self.application_path = os.path.dirname(__file__)
        self.firmware_path = os.path.join(os.getcwd(), self.firmware_defaul_filename)
        self.get_uarm_port()

    def get_uarm_port(self):
        uarm_list = uarm_ports()
        if len(uarm_list) > 0:
            self.uarm_port = uarm_list[0]
        else:
            print "No uArm is connected."

    def flash_firmware(self, firmware_path='firmware.hex'):
        global avrdude_path, error_description, cmd
        port = '-P' + self.uarm_port
        if platform.system() == 'Darwin':
            avrdude_path = 'avrdude'
            # conf_name = 'avrdude.conf'
            cmd = [avrdude_path, '-v', '-patmega328p', '-carduino', port, '-b115200', '-D',
               '-Uflash:w:{0}:i'.format(firmware_path)]
            error_description = "avrdude is required, Please try `brew install avrdude`"
        elif platform.system() == 'Windows':
            avrdude_path = os.path.join(self.application_path, 'avrdude', 'avrdude.exe')
            conf_name = 'avrdude.conf'
            avrdude_conf = os.path.join(self.application_path, 'avrdude', conf_name)
            cmd = [avrdude_path,'-C'+avrdude_conf, '-v', '-patmega328p', '-carduino', port, '-b115200', '-D',
               '-Uflash:w:{0}:i'.format(firmware_path)]
        elif platform.system() == 'Linux':
            avrdude_path = 'avrdude'
            # conf_name = 'avrdude.conf'
            cmd = [avrdude_path, '-v', '-patmega328p', '-carduino', port, '-b115200', '-D',
               '-Uflash:w:{0}:i'.format(firmware_path)]
            error_description = "avrdude is required, Please try `apt-get install avrdude`"

        print (' '.join(cmd))
        try:
            subprocess.call(cmd)
        except OSError:
            print error_description

    def get_download_url(self, release_url=github_release_url):
        url = release_url
        c = pycurl.Curl()
        data = BytesIO()
        c.setopt(pycurl.CAINFO, certifi.where())
        c.setopt(c.URL, url)
        c.setopt(c.WRITEFUNCTION, data.write)
        c.perform()
        dict = json.loads(data.getvalue())
        firmware_url = dict['assets'][0]['browser_download_url']
        firmware_size = dict['assets'][0]['size']
        self.firmware_url = firmware_url
        self.firmware_size = firmware_size
        print ("Firmware URL: {0}".format(firmware_url))
        print ("Firmware Size: {0}".format(firmware_size))

    def download_firmware(self):
        print ("Downloading firmware.hex...")
        self.get_download_url()
        try:
            response = requests.get(self.firmware_url, stream=True)
            with open(self.firmware_path, "wb") as handle:
                widgets = ['Downloading: ', Percentage(), ' ',
                           Bar(marker='#', left='[', right=']'),
                           ' ', ETA(), ' ', FileTransferSpeed()]
                pbar = ProgressBar(widgets=widgets, maxval=self.firmware_size)
                pbar.start()
                for i, data in izip(range(self.firmware_size), response.iter_content()):
                    handle.write(data)
                    pbar.update(i)
                pbar.finish()
        except requests.exceptions.ConnectionError:
            raise NetworkError("NetWork Error, Please retry...")

    def get_latest_version(self, release_url=github_release_url):
        c = pycurl.Curl()
        data = BytesIO()
        c.setopt(pycurl.CAINFO, certifi.where())
        c.setopt(c.URL, release_url)
        c.setopt(c.WRITEFUNCTION, data.write)
        c.perform()
        dict = json.loads(data.getvalue())
        self.web_firmware_version = dict['tag_name']

    def get_uarm_version(self):
        if self.uarm_port is not None or self.uarm_port != "":
            print "Reading Firmware version from uArm..."
            try:
                uarm = pyuarm.uArm(port=self.uarm_port)
                self.uarm_firmware_version = uarm.firmware_version
                uarm.disconnect()
            except pyuarm.UnkwonFirmwareException:
                print "Unknown Firmware version."
            except pyuarm.NoUArmPortException:
                print "No uArm is connected."

    def comapre_version(self):
        self.get_latest_version()
        self.get_uarm_version()
        if LooseVersion(self.web_firmware_version) > LooseVersion(self.uarm_firmware_version):
            return True


def main():
    """


    ::

        $ python -m pyuarm.tools.firmware_helper -h

        usage: firmware_helper.py [-h] [-d] [-f [FORCE]] [-c [CHECK]]

        optional arguments:
          -h, --help            show this help message and exit
          -d, --download        download firmware into firmware.hex
          -f [FORCE], --force [FORCE]
                                without firmware path, flash default firmware.hex, if
                                not existed, download automatically, with firmware
                                path, flash the firmware, eg. -f Blink.ino.hex
          -c [CHECK], --check [CHECK]
                                remote - lateset firmware release version, local -
                                read uArm firmware version
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("-d", "--download", help="download firmware into firmware.hex", action="store_true")
    parser.add_argument("-f", "--force", nargs='?', const='download',
                        help="without firmware path, flash default firmware.hex, if not existed, download automatically, with firmware path, flash the firmware, eg. -f Blink.ino.hex")
    parser.add_argument("-c", "--check", nargs='?', const='local',
                        help="remote - lateset firmware release version, local - read uArm firmware version")
    args = parser.parse_args()
    helper = FirmwareHelper()
    # download
    if args.download:
        helper.download_firmware()
        sys.exit(0)
    # force
    if args.force == "download":
        if not os.path.exists(helper.firmware_path):
            print "firmware not existed, Downloading..."
            helper.download_firmware()
        helper.flash_firmware()
        sys.exit(0)
    elif args.force is not None and args.force != "":
        if not os.path.exists(args.force):
            print args.force + " not existed."
        else:
            helper.flash_firmware(firmware_path=args.force)
            exit_fun()
    # check
    if args.check == "local":
        if helper.uarm_port is None or helper.uarm_port == "":
            sys.exit(0)
        helper.get_uarm_version()
        sys.exit(0)
    elif args.check == "remote":
        print "Fetching the remote version..."
        helper.get_download_url()
        helper.get_latest_version()
        print "Latest firmware release version is: {0}".format(helper.web_firmware_version)
        sys.exit(0)

    if helper.uarm_port is None or helper.uarm_port == "":
        exit_fun()

    # No Argument#####
    try:
        version_compare = helper.comapre_version()
        print "Latest Firmware version: {0}".format(helper.web_firmware_version)
        # print "Your uArm Firmware version: {0}".format(helper.uarm_firmware_version)
        if version_compare:
            print ("Would you want to upgrade your uArm with {0}{1}".format(helper.web_firmware_version, "?"))
            user_choice = raw_input("Please Enter Y if yes. ")
            if user_choice == "Y" or user_choice == "y":
                helper.download_firmware()
                helper.flash_firmware()
            else:
                exit_fun()
        else:
            print("You already have the latest version of Firmware installed in uArm!!!")
            exit_fun()
    except Exception:
        print "Latest Firmware version: {0} ".format(helper.web_firmware_version)
        print ("Unknown uArm Firmware version, Would you want to upgrade your uArm with {0}{1}".format(helper.web_firmware_version, "?"))
        user_choice = raw_input("Please Enter Y if yes. ")
        if user_choice == "Y" or user_choice == "y":
            helper.download_firmware()
            helper.flash_firmware()
        else:
            sys.exit(0)


def exit_fun():
    raw_input("Press Enter to Exit...")
    sys.exit(0)


class NetworkError(Exception):
    def __init__(self, error):
        self.error = error

    def __str__(self):
        return repr(self.error)


if __name__ == '__main__':
    try:
        main()
    except SystemExit:
        print("Exit")



from serial.tools import list_ports

UARM_HWID_KEYWORD = "USB VID:PID=0403:6001"


def uarm_ports():
    uarm_ports = []
    for i in list_ports.comports():
        if i.hwid[0:len(UARM_HWID_KEYWORD)] == UARM_HWID_KEYWORD:
            uarm_ports.append(i[0])
    return uarm_ports


def main():
    """
    ::

        $ python -m pyuarm.tools.list_uarms
        /dev/cu.usbserial-A600CVS9
        1 ports found

    """
    ports = uarm_ports()
    for p in ports:
        print (p)
    print ("{0} ports found".format(len(ports)))

if __name__ == '__main__':
    main()



from setuptools import setup, find_packages

from distutils.util import convert_path


long_description = open('README.rst').read()

main_ns = {}
ver_path = convert_path('pyuarm/version.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(name='pyuarm',
    version=main_ns['__version__'],
    author='Joey Song/Alex Tan',
    packages=find_packages(),
    scripts=['pyuarm/tools/firmware_helper.py', 'pyuarm/tools/calibrate.py', 'pyuarm/tools/list_uarms.py'],
    package_data={'pyuarm.tools': ['avrdude/*'] },
    include_package_data=True,
    author_email='developer@ufactory.cc',
    url="https://github.com/uarm-developer/pyuarm",
    keywords="pyuarm uarm4py uarmForPython uarm ufactory",
    install_requires=requirements,
    long_description=long_description,
    description='A python library for uArm',
)



from __future__ import print_function
import serial
import time
import binascii
from util import *


class uArm(object):
    """
    **uArm** class is let you make an instance for uArm.

    eg.

    ::

        >>> import pyuarm
        >>> uarm = pyuarm.uArm()
        Initialize uArm, port is /dev/cu.usbserial-A600CVS1...
        Firmware Version: 1.5.11

    :param port: uArm Port name, Default is **None**

    :param debug: Debug Mode Control Flag, If debug is **True**, Will display all in/out serial message(in Hex Format)

    :param timeout: Serial Timeout parameter, Default is **5** Seconds.

    :raise NoUArmPortException: No uArm available
    :raise UnkwonFirmwareException: Not unrecognized firmware version
    """

    def __init__(self, port=None, debug=False, timeout=5):

        if port is None:
            ports = uarm_ports()
            if len(ports) > 0:
                port = ports[0]
            else:
                raise NoUArmPortException("No uArm is connected.")
        self.port = port
        self.debug = debug
        print("Initialize uArm, port is {0}...".format(self.port))
        self.sp = serial.Serial(port, baudrate=57600, timeout=timeout)
        time.sleep(3)
        self.get_firmware_version()
        print("Firmware Version: {0}".format(self.firmware_version))
        # try:
        #     # self.get_firmata_version()
        #     self.get_firmware_version()
        # except serial.SerialException as e:
        #     raise UnkwonFirmwareException(
        #         "Unkwon Firmware Version, Please use 'pyuarm.tools.firmware_helper' upgrade your firmware")
        # except TypeError as e:
        #     raise UnkwonFirmwareException(
        #         "Unkwon Firmware Version, Please use 'pyuarm.tools.firmware_helper' upgrade your firmware")
        # print("Firmware Version: {0}".format(self.firmware_version))

    def is_connected(self):
        """
        This function will detect if uArm is connected

        :returns: True - Connected, False - Disconnected
        """
        if not self.sp.isOpen():
            return False
        else:
            return True

    def disconnect(self):
        """
        Close serial connection and release the uArm Port.
        """
        self.sp.close()

    def reconnect(self):
        """
        Reconnect serial connection with uArm Port.
        """
        if not self.is_connected():
            self.sp.open()
            self.get_firmata_version()
            self.get_firmware_version()

    def get_firmata_version(self):
        """
        Read frimata version after initialize the uArm.
        """
        while self.sp.readable():
            read_byte = self.serial_read()
            if read_byte == END_SYSEX:
                break
            if read_byte == START_SYSEX:
                read_byte = self.serial_read()
                if read_byte == REPORT_FIRMATA_VERSION:
                    frimata_major_version = self.serial_read()
                    frimata_minor_version = self.serial_read()
                    self.firmata_version = str(frimata_major_version) + "." + str(frimata_minor_version)

    def read_servo_angle(self, servo_number=None, with_offset=True):
        """
        Read Servo Angle from servo_number,
        if servo_number is None, will return all servo angles (SERVO_BOTTOM, SERVO_LEFT, SERVO_RIGHT, SERVO_HAND).

        eg.

        ::

            >>> import pyuarm
            >>> uarm = pyuarm.uArm(debug=True)
            Initialize uArm, port is /dev/cu.usbserial-A600CRCD...
            f925Serial Read: f079135504107206d04606907206d06107406102e06906e06f0f7
            Serial Write:f0aa23f7
            Serial Read: f0aa2315bf7
            Firmware Version: 1.5.11
            >>> uarm.read_servo_angle(pyuarm.SERVO_BOTTOM)
            Serial Write:f0aa100001f7
            Serial Read: f0aa05c034f7
            92.52
            >>> uarm.read_servo_angle(pyuarm.SERVO_BOTTOM,False)
            Serial Write:f0aa100000f7
            Serial Read: f0aa058015f7
            88.21
            >>> uarm.read_servo_angle()
            Serial Write: f0aa100001f7
            Serial Read: f0aa00270016f7
            Serial Write: f0aa100101f7
            Serial Read: f0aa016a0037f7
            Serial Write: f0aa100201f7
            Serial Read: f0aa02140000f7
            Serial Write: f0aa100301f7
            Serial Read: f0aa0306004bf7
            [39.22, 106.55, 20.0, 6.75]

        :param servo_number:    SERVO_BOTTOM
                                SERVO_LEFT
                                SERVO_RIGHT
                                SERVO_RIGHT
        :param with_offset: default - True

        """
        if servo_number is not None:

            msg = bytearray([START_SYSEX, UARM_CODE, READ_ANGLE])
            msg.extend(getOne7BitBytesFloatArray(servo_number))

            msg.extend(getOne7BitBytesFloatArray(1 if with_offset else 0))
            msg.append(END_SYSEX)
            self.serial_write(msg)
            while self.sp.readable():
                read_byte = self.serial_read()
                received_data = []
                if read_byte == START_SYSEX:
                    read_byte = self.serial_read()
                    if read_byte == UARM_CODE:
                        read_byte = self.serial_read()
                        while read_byte != END_SYSEX:
                            received_data.append(read_byte)
                            read_byte = self.serial_read()
                        if received_data[0] == servo_number:
                            return received_data[2] * 128 + received_data[1] + received_data[3] / 100.00
        else:
            servo_angles = [self.read_servo_angle(SERVO_BOTTOM, with_offset),
                            self.read_servo_angle(SERVO_LEFT, with_offset),
                            self.read_servo_angle(SERVO_RIGHT, with_offset),
                            self.read_servo_angle(SERVO_HAND, with_offset)]
            return servo_angles


    def read_eeprom(self, data_type, eeprom_address):
        """
        Read data from EEPROM

        Byte: 1 Byte

        Integer: 2 Bytes

        Float: 4 Bytes

        Please reference Arduino EEPROMGet_.

        .. _EEPROMGet: https://www.arduino.cc/en/Reference/EEPROMGet

        :param data_type:   EEPROM_DATA_TYPE_BYTE

                            EEPROM_DATA_TYPE_FLOAT

                            EEPROM_DATA_TYPE_INTEGER

        :param eeprom_address: EEPROM Address
        :returns: byte, integer, float
        """
        msg = bytearray([START_SYSEX, UARM_CODE, READ_EEPROM])
        msg.append(data_type)
        msg.extend(getTwo7BitBytesIntegerArray(eeprom_address))
        msg.append(END_SYSEX)

        self.serial_write(msg)
        while self.sp.readable():
            read_byte = self.serial_read()
            received_data = []
            if read_byte == START_SYSEX:
                read_byte = self.serial_read()
                if read_byte == UARM_CODE:
                    read_byte = self.serial_read()
                    if read_byte == READ_EEPROM:
                        read_byte = self.serial_read()
                        while read_byte != END_SYSEX:
                            received_data.append(read_byte)
                            read_byte = self.serial_read()
                        res_eeprom_add = received_data[1] * 128 + received_data[0]
                        if res_eeprom_add == eeprom_address:
                            if data_type == EEPROM_DATA_TYPE_BYTE:
                                return (received_data[3] << 7) + received_data[2]
                            if data_type == EEPROM_DATA_TYPE_FLOAT:
                                val = (received_data[4] << 7) + received_data[3] + float(received_data[5]) / 100.00
                                val = val if received_data[2] == 0 else -val
                                return val
                            if data_type == EEPROM_DATA_TYPE_INTEGER:
                                val = (received_data[4] << 7) + received_data[3]
                                val = val if received_data[2] == 0 else -val
                                return val

    def read_digital(self, pin_num, pin_mode):
        """
        Read Digital from PIN

        Please reference Arduino DigitalPINs_.

        .. _DigitalPINs: https://www.arduino.cc/en/Tutorial/DigitalPins

        :param pin_num: Integer 1 to 13
        :param pin_mode: PULL_UP, INPUT
        :returns: Digital Value - HIGH, LOW

        """
        msg = bytearray([START_SYSEX, UARM_CODE, READ_DIGITAL])
        msg.extend(getOne7BitBytesFloatArray(pin_num))
        msg.extend(getOne7BitBytesFloatArray(pin_mode))
        msg.append(END_SYSEX)

        self.serial_write(msg)
        while self.sp.readable():
            read_byte = self.serial_read()
            received_data = []
            if read_byte == START_SYSEX:
                read_byte = self.serial_read()
                if read_byte == UARM_CODE:
                    read_byte = self.serial_read()
                    while read_byte != END_SYSEX:
                        received_data.append(read_byte)
                        read_byte = self.serial_read()
                    if received_data[0] == pin_num:
                        return received_data[1]

    def read_analog(self, pin_num):
        """
        Read Analog from PIN

        Please reference Arduino AnalogRead_.

        .. _AnalogRead: https://www.arduino.cc/en/Reference/AnalogRead

        :param pin_num: Integer 1 to 13
        :returns: Analog Value - Integer

        """
        msg = bytearray([START_SYSEX, UARM_CODE, READ_ANALOG])
        msg.extend(getOne7BitBytesFloatArray(pin_num))
        msg.append(END_SYSEX)
        self.serial_write(msg)
        while self.sp.readable():
            read_byte = self.serial_read()
            received_data = []
            if read_byte == START_SYSEX:
                read_byte = self.serial_read()
                if read_byte == UARM_CODE:
                    read_byte = self.serial_read()
                    while read_byte != END_SYSEX:
                        received_data.append(read_byte)
                        read_byte = self.serial_read()
                    if received_data[0] == pin_num:
                        return received_data[2] * 128 + received_data[1]

    def read_coordinate(self):
        """
        Read Coordinate from uArm.

        :return: Coordinate array, x, y, z, Float type

        """

        msg = bytearray([START_SYSEX, UARM_CODE, READ_COORDS])
        msg.append(END_SYSEX)
        self.serial_write(msg)

        if self.sp.readable():

            if self.serial_read() == START_SYSEX:
                if self.serial_read() == UARM_CODE:
                    if self.serial_read() == READ_COORDS:
                        coordinate = [getFloatFromFour7BitBytes(self.serial_read(4)), getFloatFromFour7BitBytes(self.serial_read(4)),
                                      getFloatFromFour7BitBytes(self.serial_read(4))] #x, y, z
                        if self.serial_read() == END_SYSEX: # END_SYSEX
                            return coordinate

    def write_eeprom(self, data_type, eeprom_add, eeprom_val):
        """
        Write Data to EEPROM

        Please reference Arduino EEPROMPut_.

        :: _EEPROMPut: https://www.arduino.cc/en/Reference/EEPROMPut

        :param data_type: EEPROM_TYPE_BYTE, EEPROM_TYPE_INTEGER, EEPROM_TYPE_FLOAT
        :param eeprom_add: EEPROM address Integer
        :param eeprom_val: EEPROM Value - Byte, Integer, Float

        """
        msg = bytearray([START_SYSEX, UARM_CODE, WRITE_EEPROM, data_type])
        msg.extend(getTwo7BitBytesIntegerArray(eeprom_add))
        if data_type == EEPROM_DATA_TYPE_BYTE:
            msg.extend(getTwo7BitBytesIntegerArray(eeprom_val))
        if data_type == EEPROM_DATA_TYPE_FLOAT:
            msg.extend(getFour7BitBytesFloatArray(eeprom_val))
        if data_type == EEPROM_DATA_TYPE_INTEGER:
            msg.extend(getThree7BitBytesIntegerArray(eeprom_val))
        msg.append(END_SYSEX)
        self.serial_write(msg)

    def write_analog(self, pin_number, pin_value):
        """
        Write Analog to PIN

        Please reference Arduino AnalogWrite_.

        :: _AnalogWrite: https://www.arduino.cc/en/Reference/AnalogWrite
        :param pin_number: PIN Number Integer 0 - 13
        :param pin_value: Analog Value

        """
        msg = bytearray([START_SYSEX, UARM_CODE, WRITE_ANALOG])
        msg.extend(getOne7BitBytesFloatArray(pin_number))
        msg.extend(getTwo7BitBytesIntegerArray(pin_value))
        msg.append(END_SYSEX)
        self.serial_write(msg)

    def write_digital(self, pin_number, pin_mode):
        """
        Write Digital to PIN

        Please reference Arduino DigitalWrite_.

        :: _DigitalWrite: https://www.arduino.cc/en/Reference/DigitalWrite

        :param pin_number: PIN Number - Integer 0~13
        :param pin_mode: Digital Value - HIGH, LOW

        """
        msg = bytearray([START_SYSEX, UARM_CODE, WRITE_DIGITAL])
        msg.extend(getOne7BitBytesFloatArray(pin_number))
        msg.extend(getOne7BitBytesFloatArray(pin_mode))
        msg.append(END_SYSEX)
        self.serial_write(msg)

    def detach_all_servos(self):
        """
        Detach All Servos

        Unlock the servo after detach.
        """
        msg = bytearray([START_SYSEX, UARM_CODE, DETACH_SERVO, END_SYSEX])
        self.serial_write(msg)

    def write_servo_angle(self, servo_number, servo_angle, with_offset):
        """
        Write Servo Angle

        :param servo_number: SERVO_BUTTON_NUM, SERVO_LEFT_NUM, SERVO_RIGHT_NUM, SERVO_HAND_NUM
        :param servo_angle: float type between 0.00 ~ 180.00
        :param with_offset: True, False

        """
        servo_number = int(servo_number)
        servo_angle = float(servo_angle)
        with_offset = int(with_offset)
        msg = bytearray([START_SYSEX, UARM_CODE, WRITE_ANGLE])
        msg.extend(getOne7BitBytesFloatArray(servo_number))
        msg.extend(getThree7BitBytesFloatArray(servo_angle))
        msg.extend(getOne7BitBytesFloatArray(with_offset))
        msg.append(END_SYSEX)
        self.serial_write(msg)

    def write_left_right_servo_angle(self, servo_left_angle, servo_right_angle, with_offset):
        """

        :param servo_left_angle:
        :param servo_right_angle:
        :param with_offset:

        """
        with_offset = int(with_offset)
        servo_left_angle = float(servo_left_angle)
        servo_right_angle = float(servo_right_angle)
        msg = bytearray([START_SYSEX, UARM_CODE, WRITE_LEFT_RIGHT_ANGLE])
        msg.extend(getThree7BitBytesFloatArray(servo_left_angle))
        msg.extend(getThree7BitBytesFloatArray(servo_right_angle))
        msg.extend(getOne7BitBytesFloatArray(with_offset))
        msg.append(END_SYSEX)
        self.serial_write(msg)

    def move_to(self, x, y, z, hand_angle=0, relative_flags=ABSOLUTE, time_spend=2, path_type=PATH_LINEAR, ease_type=INTERP_EASE_INOUT_CUBIC):
        """
        Move uArm to a `coordinate(x,y,z)` or `Angles(bottom,left,right,hand)`.

        Default parameters `move_to(x,y,z,0, ABSOLUTE, 2, PATH_LINEAR, INTERP_EASE_INOUT_CUBIC)`

        Move uArm to a `coordinate(x,y,z)` in 2 seconds.

        :param x: Float type, X Axis.
        :param y: Float type, Y Axis.
        :param z: Float type, Z Axis.
        :param hand_angle: Float Type, Hand Angle.
        :param relative_flags: ABSOLUTE, RELATIVE, default is ABSOLUTE
        :param time_spend: Float type, time_spend, default is 2
        :param path_type: Path Type, PATH_LINEAR, PATH_ANGLES, default is PATH_LINEAR
        :param ease_type: Ease Type, INTERP_EASE_INOUT_CUBIC, INTERP_LINEAR, INTERP_EASE_INOUT, INTERP_EASE_IN, INTERP_EASE_OUT

        """
        x, y, z = float(x), float(y), float(z)
        msg = bytearray([START_SYSEX, UARM_CODE, WRITE_COORDS])
        msg.extend(getFour7BitBytesFloatArray(x))
        msg.extend(getFour7BitBytesFloatArray(y))
        msg.extend(getFour7BitBytesFloatArray(z))
        msg.extend(getThree7BitBytesFloatArray(hand_angle))
        msg.extend(getOne7BitBytesFloatArray(relative_flags))
        msg.extend(getThree7BitBytesFloatArray(time_spend))
        msg.extend(getOne7BitBytesFloatArray(path_type))
        msg.extend(getOne7BitBytesFloatArray(ease_type))
        msg.append(END_SYSEX)
        time.sleep(0.01)
        self.serial_write(msg)

    def pump_control(self, val):
        """
        Control Pump On or Off
        :param val: True is On, False is Off.

        """
        pump_status = 1 if val else 0
        msg = bytearray([START_SYSEX, UARM_CODE, PUMP_STATUS])
        msg.extend(getOne7BitBytesFloatArray(pump_status))
        msg.append(END_SYSEX)
        self.serial_write(msg)

    def gripper_control(self, val):
        """
        Control Gripper On or Off
        :param val: True is On, False is Off.

        """
        gripper_status = 1 if val else 0
        msg = bytearray([START_SYSEX, UARM_CODE, GRIPPER_STATUS])
        msg.extend(getOne7BitBytesFloatArray(gripper_status))
        msg.append(END_SYSEX)
        self.serial_write(msg)

    def write_stretch(self, length, height):
        """
        Write Stretch is another control method for uArm.

        Length means uArm Y Axis. between (0, 195)
        Height means uArm Z Axis. between (-150, 130)

        :param length: Float Type, between (0, 195)
        :param height: Float Type, between (-150, 130)

        """
        length = float(length)
        height = float(height)
        msg = bytearray([START_SYSEX, UARM_CODE, WRITE_STRETCH])
        msg.extend(getFour7BitBytesFloatArray(length))
        msg.extend(getFour7BitBytesFloatArray(height))
        msg.append(END_SYSEX)
        self.serial_write(msg)

    def write_serial_number(self, serial_number):
        """
        Write Serial Number to uArm.
        Serial Number is 15 digital number
        eg. UARM030715011B
        :param serial_number:

        """
        msg = bytearray([START_SYSEX, UARM_CODE, WRITE_SERIAL_NUMBER])
        for c in serial_number:
            msg.append(ord(c))
        msg.append(END_SYSEX)
        self.serial_write(msg)

    def read_serial_number(self):
        """
        Read Serial Number from uArm.
        Serial Number is 15 digital number
        eg. UARM030715011B
        """
        msg = bytearray([START_SYSEX, UARM_CODE, READ_SERIAL_NUMBER, END_SYSEX])
        self.serial_write(msg)
        while self.sp.readable():
            read_byte = self.serial_read()
            received_data = []
            if read_byte == START_SYSEX:
                read_byte = self.serial_read()
                if read_byte == UARM_CODE:
                    read_byte = self.serial_read()
                    if read_byte == READ_SERIAL_NUMBER:
                        read_byte = self.serial_read()
                        while read_byte != END_SYSEX:
                            received_data.append(read_byte)
                            read_byte = self.serial_read()
                        sn_array = []
                        for r in received_data:
                            sn_array.append(chr(r))
                        return ''.join(sn_array)

    def alarm(self, times, run_time, stop_time):
        """
        Alarm the Buzzer.
        eg.

        ::

            >>> import pyuarm
            >>> uarm = pyuarm.uArm(debug=True)
            Initialize uArm, port is /dev/cu.usbserial-A600CVS9...
            f925Serial Read: f079135504107206d04606907206d06107406102e06906e06f0f7
            Serial Write:f0aa23f7
            Serial Read: f0aa2315bf7
            Firmware Version: 1.5.11
            >>> uarm.alarm(1, 100, 100)
            Serial Write:f0aa24016464f7
            >>> uarm.alarm(3, 100, 100)
            Serial Write:f0aa24036464f7


        :param times: Alarm times
        :param run_time: Alarm Run Times.
        :param stop_time: Alarm Stop Times.

        """
        msg = bytearray([START_SYSEX, UARM_CODE, BUZZER_ALERT, times, run_time, stop_time, END_SYSEX])
        self.serial_write(msg)

    def get_firmware_version(self):
        """
        Read Firmware Vresion from uArm.
        """
        if self.is_connected():
            msg = bytearray([START_SYSEX, UARM_CODE, REPORT_FIRMWARE_VERSION, END_SYSEX])
            self.serial_write(msg)
            while self.sp.readable():
                read_byte = self.serial_read()
                received_data = []
                if read_byte == START_SYSEX:
                    read_byte = self.serial_read()
                    if read_byte == UARM_CODE:
                        read_byte = self.serial_read()
                        if read_byte == REPORT_FIRMWARE_VERSION:
                            read_byte = self.serial_read()
                            while read_byte != END_SYSEX:
                                received_data.append(read_byte)
                                read_byte = self.serial_read()
                            firmware_major_version = received_data[0]
                            firmware_minor_version = received_data[1]
                            firmware_bugfix = received_data[2]
                            self.firmware_version = str(firmware_major_version) + "." + str(
                                firmware_minor_version) + "." + str(firmware_bugfix)
                            break
        else:
            print("uArm is not Connected")

    def serial_write(self, msg):
        """
        Write message to Serial Port.
        If debug is True, will display all Serial Write message.

        :param msg: Serial Message, bytearray.

        """
        if self.debug:
            print("Serial Write: {0}".format(binascii.hexlify(msg)))
        self.sp.write(msg)

    def serial_read(self, count=1):
        """
        Read one byte from Serial Port. If Debug is True, will display all Serial Read message.
        """
        if count == 1:
            read_byte = ord(self.sp.read(1))
            if self.debug:
                if read_byte == START_SYSEX:
                    print ("Serial Read:", end=' ')
                print(format(read_byte, '02x'), end='')
                if read_byte == END_SYSEX:
                    print("")
            return read_byte
        else:
            read_bytes = self.sp.read(count)
            if self.debug:
                print(binascii.hexlify(read_bytes), end='')
            return read_bytes
